﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading;
using CommandLine;
using System.IO;
using System.ComponentModel.DataAnnotations;
using System.Globalization;
using System.Runtime.InteropServices;
using System.Security.Cryptography.X509Certificates;
using System.Collections;
using System.Drawing;
using System.Runtime;
using System.Runtime.CompilerServices;

namespace files_to_shm
{
    class Program
    {
		public struct options
		{
			public string baseDir;
			public string csv;
			public string ext;
			public int maxImages;
			public string cam0;
			public string cam1;
			public string cam2;
			public string cam3;
			public string cam4;
			public string cam5;
			public string lidar;
			public bool verbose;
			public int delay;
            public int off0;
            public int off1;
            public int off2;
            public int off3;
            public int off4;
            public int off5;
        }

		//some static stuff for LiDAR pointclouds
		public static int requiredPointNumber = 14976;
		public static float[] singleFrameArray = new float[requiredPointNumber * 4]; //big array that stores a single full pointcloud
		public static int max_memory_size = singleFrameArray.Length * sizeof(float); //init shared memory size

		//initializing shared memory for sync 1/2
		//init shared memory for LiDAR
		public static string sync_memory_name = "sync_mem";
		public static System.IO.MemoryMappedFiles.MemoryMappedFile sync_shared_memory = System.IO.MemoryMappedFiles.MemoryMappedFile.CreateOrOpen(sync_memory_name, sizeof(int));

		public static string[] memory_names = { "shm_1", "shm_2" };
		public static int[] memory_sizes = { max_memory_size, max_memory_size };
		public static System.IO.MemoryMappedFiles.MemoryMappedFile[] shared_memories = new System.IO.MemoryMappedFiles.MemoryMappedFile[2];


		//init shared memories for cam sync
		public static string camera_sync_name = "camera_flag";
		public static System.IO.MemoryMappedFiles.MemoryMappedFile syncCamera_memory = System.IO.MemoryMappedFiles.MemoryMappedFile.CreateOrOpen(camera_sync_name, sizeof(int));

		public static string[] camera_memory_names = { "cameraData_dev0seq1", "cameraData_dev1seq1", "cameraData_dev2seq1", "cameraData_dev3seq1",
			"cameraData_dev4seq1", "cameraData_dev5seq1", "cameraData_dev0seq2", "cameraData_dev1seq2", "cameraData_dev2seq2", "cameraData_dev3seq2",
			"cameraData_dev4seq2", "cameraData_dev5seq2" };
		public static int camera_frame_size = 6914048;
		public static byte[] newFrame = new byte[camera_frame_size];
		public static System.IO.MemoryMappedFiles.MemoryMappedFile[] shared_camera_frames = new System.IO.MemoryMappedFiles.MemoryMappedFile[12];


		//init shared memory for meadata sync
		public static string meta_sync_name = "meta_flag";
		public static System.IO.MemoryMappedFiles.MemoryMappedFile syncMetaMemory=System.IO.MemoryMappedFiles.MemoryMappedFile.CreateOrOpen(meta_sync_name,sizeof(int));

        public static string[] meta_memory_names = { "meta_0", "meta_1" };
        public static int meta_sync_size = 11 * sizeof(float);
		public static System.IO.MemoryMappedFiles.MemoryMappedFile[] shared_metadata = new System.IO.MemoryMappedFiles.MemoryMappedFile[2];


        //not so elegant globals
        public static options runParams = new options();
		public static int active_buffer_flag = 0; //active memory flag for cameras
		public static byte[] nagyonNull = new byte[camera_frame_size]; //fixed size null buffer to insert when a file does not exist
		public static int fn = 1; //current frame number counter
		public static string specifier = "G"; //text formatting bullshit
		public static CultureInfo culture = CultureInfo.CreateSpecificCulture("en-US");//text formatting bullshit
		public static bool isPaused = false;

		public static CsvHandler metaHandler = new CsvHandler();


		static void Main(string[] args)
		{
			ConsoleKeyInfo cki;
			do
			{
                // Initializing shared memories for the data !!2/2!!
                for (int i = 0; i < memory_names.Length; i++)
                {
                    shared_memories[i] = System.IO.MemoryMappedFiles.MemoryMappedFile.CreateOrOpen(memory_names[i], memory_sizes[i]);
                }
                for (int i = 0; i < camera_memory_names.Length; ++i)
                {
                    shared_camera_frames[i] = System.IO.MemoryMappedFiles.MemoryMappedFile.CreateOrOpen(camera_memory_names[i], camera_frame_size);
                }
				for (int i = 0;i<meta_memory_names.Length;i++)
				{
					shared_metadata[i] = System.IO.MemoryMappedFiles.MemoryMappedFile.CreateOrOpen(meta_memory_names[i], meta_sync_size);
				}
                CommandLine.Parser.Default.ParseArguments<Args>(args).WithParsed(ArgParse).WithNotParsed(HandleParseError);
                runParams.baseDir = runParams.baseDir + "\\";
                metaHandler = new CsvHandler();

                metaHandler.loadcsv(runParams.csv);

                //FPS timer (default is 4 FPS, to meet with lidar_to_shm)
                Timer fpstick = new Timer(TimerCallback, null, 0, runParams.delay);

                Console.WriteLine("press R to reset!");
				Console.WriteLine("press P to pause");
                Console.WriteLine("PRESS ENTER TO EXIT!");
                Console.WriteLine("reading files to memory...");
				cki = Console.ReadKey();
				if(cki.Key==ConsoleKey.R)
				{
					fn = 1;
				}
				if(cki.Key==ConsoleKey.P)
				{
					if(isPaused)
					{
						isPaused = false;
						Console.WriteLine("Resume");
					}
					else
					{
						isPaused = true;
						Console.WriteLine("Pause");
					}
				}
            }
			while(cki.Key!=ConsoleKey.Enter);
        }

		public static void TimerCallback(Object o)
		{
			GCLatencyMode oldMode = GCSettings.LatencyMode;
			RuntimeHelpers.PrepareConstrainedRegions();
			try
			{
				GCSettings.LatencyMode = GCLatencyMode.LowLatency;

                int id = 0;
                if (runParams.cam0 != "not defined")
                {
                    id = 0;
                    fillCameras(id);
                }
                if (runParams.cam1 != "not defined")
                {
                    id = 1;
                    fillCameras(id);
                }
                if (runParams.cam2 != "not defined")
                {
                    id = 2;
                    fillCameras(id);
                }
                if (runParams.cam3 != "not defined")
                {
                    id = 3;
                    fillCameras(id);
                }
                if (runParams.cam4 != "not defined")
                {
                    id = 4;
                    fillCameras(id);
                }
                if (runParams.cam5 != "not defined")
                {
                    id = 5;
                    fillCameras(id);
                }
                if (runParams.lidar != "not defined")
                {
                    fillLidar();
                }
                if (runParams.csv != "not defined")
                {
                    fillMetaData();
                }
                if (isPaused == false)
                {
                    fn++;
                }
                if (fn > runParams.maxImages)
                {
                    Console.WriteLine("Ran out of frames provided, starting from the beginning");
                    fn = 1;
                }


                //update syncCamera_memory to match active_buffer_flag
                System.IO.MemoryMappedFiles.MemoryMappedViewStream update_sync_view;
                update_sync_view = syncCamera_memory.CreateViewStream();
                byte[] syncBytes = BitConverter.GetBytes(active_buffer_flag);
                update_sync_view.Write(syncBytes, 0, sizeof(int));
                update_sync_view.Close();

                //update sync_shared_memory to match active_buffer_flag
                update_sync_view = sync_shared_memory.CreateViewStream();
                syncBytes = BitConverter.GetBytes(active_buffer_flag);
                update_sync_view.Write(syncBytes, 0, sizeof(int));
                update_sync_view.Close();

                //update meta_flag to match active_buffer_flag
                update_sync_view = syncMetaMemory.CreateViewStream();
                syncBytes = BitConverter.GetBytes(active_buffer_flag);
                update_sync_view.Write(syncBytes, 0, sizeof(int));
                update_sync_view.Close();

                //updating buffer flag
                active_buffer_flag++;
                if (active_buffer_flag > 1)
                {
                    active_buffer_flag = 0;
                }
            }
			finally
			{
				GCSettings.LatencyMode = oldMode;
			}
			GC.Collect();

		}

		static void fillCameras(int id)
		{
			string fullpath = "";
			System.IO.MemoryMappedFiles.MemoryMappedViewStream view;
			int num = 0;

			switch (id)
			{
				/*default:
					num = fn + runParams.off0;
					fullpath = runParams.baseDir + runParams.cam0 + num.ToString() + "." + runParams.ext;
					break;*/
				case 0:
                    num = fn + runParams.off0;
                    fullpath = runParams.baseDir + runParams.cam0 + num.ToString() + "." + runParams.ext;
                    break;
				case 1:
                    num = fn + runParams.off1;
                    fullpath = runParams.baseDir + runParams.cam1 + num.ToString() + "." + runParams.ext;
                    break;
				case 2:
                    num = fn + runParams.off2;
                    fullpath = runParams.baseDir + runParams.cam2 + num.ToString() + "." + runParams.ext;
                    break;
				case 3:
                    num = fn + runParams.off3;
                    fullpath = runParams.baseDir + runParams.cam3 + num.ToString() + "." + runParams.ext;
                    break;
				case 4:
                    num = fn + runParams.off4;
                    fullpath = runParams.baseDir + runParams.cam4 + num.ToString() + "." + runParams.ext;
                    break;
				case 5:
                    num = fn + runParams.off5;
                    fullpath = runParams.baseDir + runParams.cam5 + num.ToString() + "." + runParams.ext;
                    break;
			}

			if (active_buffer_flag == 0)
			{
				view = shared_camera_frames[id].CreateViewStream();
				if (runParams.verbose)
				{
					Console.WriteLine(fullpath + " to shared mem id: " + id);
				}
			}
			else
			{
				view = shared_camera_frames[id + 6].CreateViewStream();
				if (runParams.verbose)
				{
					Console.WriteLine(fullpath + " to shared mem id: " + (id + 6));
				}
			}
			if (File.Exists(fullpath))
			{
				//copy frame data from disk to a byte array, then put the byte array to shared memory
				/*FileStream fileIn = File.OpenRead(fullpath);
				fileIn.Read(newFrame, 0, newFrame.Length);
				view.Write(newFrame, 0, newFrame.Length);
				view.Close();*/

				Image image=Image.FromFile(fullpath);
				var ms=new MemoryStream();
				image.Save(ms,System.Drawing.Imaging.ImageFormat.Bmp);
				newFrame = ms.ToArray();
                view.Write(newFrame, 0, newFrame.Length);
                view.Close();

            }
			else
			{
				//handle non existing files
				view.Write(nagyonNull, 0, nagyonNull.Length);
				Console.WriteLine(fullpath + " does not exist");
			}
		}

        static void fillLidar()
        {
			string full_input_path = runParams.lidar + fn.ToString() + ".xyz";

			if(File.Exists(full_input_path))
            {
				StreamReader fileIn = File.OpenText(full_input_path);
				int singleFrameCounter = 0;
				while (!fileIn.EndOfStream)
				{
					string cLineIn = fileIn.ReadLine();
					string[] cLineSplit = cLineIn.Split(' ');

					singleFrameArray[singleFrameCounter] = float.Parse(cLineSplit[0], culture);
					singleFrameCounter++;
					singleFrameArray[singleFrameCounter] = float.Parse(cLineSplit[1], culture);
					singleFrameCounter++;
					singleFrameArray[singleFrameCounter] = float.Parse(cLineSplit[2], culture);
					singleFrameCounter++;
					singleFrameArray[singleFrameCounter] = float.Parse(cLineSplit[3], culture);
					singleFrameCounter++;
				}
				fileIn.Close();
				if (runParams.verbose)
				{
					Console.WriteLine(full_input_path + " to LiDAR shared memory ID: " + active_buffer_flag.ToString());
				}
			}
            else
            {
				for(int i=0;i>singleFrameArray.Length;i++)
                {
					singleFrameArray[i] = 0;
                }
				Console.WriteLine(full_input_path + " does not exist, filling buffer with zeroes!");
            }

			//Select buffer and fill it with the newest lidar data
			System.IO.MemoryMappedFiles.MemoryMappedViewStream view;
			view = shared_memories[active_buffer_flag].CreateViewStream();
			byte[] byteArray = new byte[max_memory_size];
			Buffer.BlockCopy(singleFrameArray, 0, byteArray, 0, byteArray.Length);
			//write converted float array to shared memory
			view.Write(byteArray, 0, byteArray.Length);
			view.Close();
		}

		static void fillMetaData()
		{
			metaLine currentMetaData = metaHandler.getMetaData(fn);
			if(currentMetaData.frameID==0&&fn!=0)
			{
				Console.WriteLine("No metadata for frameID: " + fn.ToString());
			}
			else
			{
				//convert metaLine to float array
				float[] floats = { currentMetaData.frameID, currentMetaData.lat, currentMetaData.lon, currentMetaData.alt, currentMetaData.vel, currentMetaData.Ax, currentMetaData.Ay, currentMetaData.Az, currentMetaData.Mx, currentMetaData.Ay, currentMetaData.Az };
                //select buffer and fill metadata
                System.IO.MemoryMappedFiles.MemoryMappedViewStream view;
				view = shared_metadata[active_buffer_flag].CreateViewStream();
				byte[] byteArray = new byte[meta_sync_size];
				Buffer.BlockCopy(floats,0,byteArray,0,byteArray.Length);
                //write converted float array to shared memory
                view.Write(byteArray, 0, byteArray.Length);
                view.Close();
            }
		}

		static void ArgParse(Args opts)
		{
			try
			{
				var validationContext = new ValidationContext(opts, serviceProvider: null, items: null);
				var validationResults = new List<ValidationResult>();

				var isValid = Validator.TryValidateObject(opts, validationContext, validationResults,
					validateAllProperties: true);

				if (!isValid)
				{
					foreach (var validationResult in validationResults)
					{
						Console.WriteLine(validationResult.ErrorMessage);
					}

					throw ArgumentException;
				}

				Console.WriteLine("BaseDir: " + opts.BaseDir);
				Console.WriteLine("Input format" + opts.ext);
				Console.WriteLine("MaxImages: " + opts.MaxImages);
                Console.Write("Cam0: " + opts.Cam0);
                Console.WriteLine(", frame offset: " + opts.off0);
                Console.Write("Cam0: " + opts.Cam1);
                Console.WriteLine(", frame offset: " + opts.off1);
                Console.Write("Cam0: " + opts.Cam2);
                Console.WriteLine(", frame offset: " + opts.off2);
                Console.Write("Cam0: " + opts.Cam3);
                Console.WriteLine(", frame offset: " + opts.off3);
                Console.Write("Cam0: " + opts.Cam4);
                Console.WriteLine(", frame offset: " + opts.off4);
                Console.Write("Cam0: " + opts.Cam5);
                Console.WriteLine(", frame offset: " + opts.off5);

                Console.WriteLine("verbose: " + opts.Verbose.ToString());
				Console.WriteLine("Delay: " + opts.Delay);
				Console.WriteLine("LiDAR base: " + opts.lidar);
				Console.WriteLine("CSV metadata: " + opts.csv);

				//move opts to global variable (not pretty but easier to use)
				runParams.baseDir = opts.BaseDir;
				runParams.maxImages = opts.MaxImages;
				runParams.cam0 = opts.Cam0;
				runParams.cam1 = opts.Cam1;
				runParams.cam2 = opts.Cam2;
				runParams.cam3 = opts.Cam3;
				runParams.cam4 = opts.Cam4;
				runParams.cam5 = opts.Cam5;
				runParams.verbose = opts.Verbose;
				runParams.delay = opts.Delay;
				runParams.lidar = opts.lidar;
                runParams.off0 = opts.off0;
                runParams.off1 = opts.off1;
                runParams.off2 = opts.off2;
                runParams.off3 = opts.off3;
                runParams.off4 = opts.off4;
                runParams.off5 = opts.off5;
				runParams.csv = opts.csv;
				runParams.ext = opts.ext;
            }
			catch (ArgumentException ex)
			{
				//Handle errors
				Console.WriteLine("[ArgumentException] Invalid arguments!");
				Console.WriteLine(ex.Message);
				// Exit with error code 1
				Environment.Exit(1);
			}
			catch (Exception ex)
			{
				//Handle errors
				Console.WriteLine("[Exception] Invalid arguments?");
				Console.WriteLine(ex.Message);
				// Exit with error code 1
				Environment.Exit(1);
			}
		}

		private static Exception ArgumentException { get; set; }

		private static void HandleParseError(IEnumerable<Error> errs)
		{
			//Handle errors
			foreach (var error in errs)
			{
				Console.WriteLine(error.Tag);
			}
		}
	}
}
